'''
PassportEye::Util: Interface between SKImage and the PyTesseract OCR
NB: You must have the "tesseract" tool present in your path for this to work.

Author: Konstantin Tretyakov
License: MIT
'''

from pytesseract import pytesseract
from scipy.misc import imsave

def ocr(img, mrz_mode=True):
    """Runs Tesseract on a given image. Writes an intermediate tempfile and then runs the tesseract command on the image.

    This is a simplified modification of image_to_string from PyTesseract, which is adapted to SKImage rather than PIL.

    In principle we could have reimplemented it just as well - there are some apparent bugs in PyTesseract (e.g. it
    may lose the NamedTemporaryFile due to its auto-delete behaviour).

    :param mrz_mode: when this is True (default) the tesseract is configured to recognize MRZs rather than arbitrary texts.
    """
    input_file_name = '%s.bmp' % pytesseract.tempnam()
    output_file_name_base = '%s' % pytesseract.tempnam()
    output_file_name = "%s.txt" % output_file_name_base
    try:
        imsave(input_file_name, img)

        if mrz_mode:
            config = "-psm 6 -c tessedit_char_whitelist=ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789>< -c load_system_dawg=F -c load_freq_dawg=F"
        else:
            config = None

        status, error_string = pytesseract.run_tesseract(input_file_name,
                                             output_file_name_base,
                                             lang=None,
                                             boxes=False,
                                             config=config)
        if status:
            errors = pytesseract.get_errors(error_string)
            raise pytesseract.TesseractError(status, errors)
        f = open(output_file_name)
        try:
            return f.read().strip()
        finally:
            f.close()
    finally:
        pytesseract.cleanup(input_file_name)
        pytesseract.cleanup(output_file_name)



'''
PassportEye::Util: PDF processing utilities.

Author: Konstantin Tretyakov
License: MIT
'''

## TODO: This uses PDFMiner for Python2. PDFMiner for Python3 has a slightly different interface
## hence the code might need to be updated slightly.

from pdfminer.pdfparser import PDFParser
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfinterp import PDFResourceManager
from pdfminer.pdfinterp import PDFPageInterpreter
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LTFigure, LTImage


def extract_first_jpeg_in_pdf(fstream):
    """
    Reads a given PDF file and scans for the first valid embedded JPEG image.
    Returns either None (if none found) or a string of data for the image.
    There is no 100% guarantee for this code, yet it seems to work fine with most
    scanner-produced images around.
    More testing might be needed though.

    Note that in principle there is no serious problem extracting PNGs or other image types from PDFs,
    however at the moment I do not have enough test data to try this, and the one I have seems to be unsuitable
    for PDFMiner.

    :param fstream: Readable binary stream of the PDF
    :return: String, containing the whole contents of the JPEG image or None if extraction failed.
    """
    parser = PDFParser(fstream)
    document = PDFDocument(parser)
    rsrcmgr = PDFResourceManager()
    device = PDFPageAggregator(rsrcmgr)
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    for page in PDFPage.create_pages(document):
        interpreter.process_page(page)
        layout = device.result
        for el in layout:
            if isinstance(el, LTFigure):
                for im in el:
                    if isinstance(im, LTImage):
                        # Found one!
                        st = None
                        try:
                            imdata = im.stream.get_data()
                        except:
                            # Failed to decode (seems to happen nearly always - there's probably a bug in PDFMiner), oh well...
                            imdata = im.stream.get_rawdata()
                        if imdata is not None and imdata.startswith('\xff\xd8\xff\xe0'):
                            return imdata

    return None


'''
PassportEye::MRZ: Machine-readable zone extraction and parsing.
Command-line scripts

Author: Konstantin Tretyakov
License: MIT
'''
import argparse, time, glob, pkg_resources, os, multiprocessing, logging, json, shutil
from collections import Counter
from skimage import io
import passporteye
from .image import read_mrz

def process_file(params):
    """
    Processes a file and returns the parsed MRZ (or None if no candidate regions were even found).
    """
    tic = time.time()
    filename, save_roi = params
    try:
        mrz = read_mrz(filename, save_roi=save_roi)
    except Exception:
        mrz = None
    walltime = time.time() - tic
    return (filename, mrz, walltime)

def evaluate_mrz():
    """
    A script for evaluating the current MRZ recognition pipeline by applying it to a list of files in a directory and reporting how well it went.
    """
    parser = argparse.ArgumentParser(description='Run the MRZ OCR recognition algorithm on the sample test data, reporting the quality summary.')
    parser.add_argument('-j', '--jobs', default=1, type=int, help='Number of parallel jobs to run')
    parser.add_argument('-dd', '--data-dir', default=pkg_resources.resource_filename('passporteye.mrz', 'testdata'),
                                help='Read files from this directory instead of the package test files')
    parser.add_argument('-sd', '--success-dir', default=None,
                                help='Copy files with successful (nonzero score) extraction results to this directory')
    parser.add_argument('-fd', '--fail-dir', default=None,
                                help='Copy files with unsuccessful (zero score) extraction resutls to this directory')
    parser.add_argument('-rd', '--roi-dir', default=None,
                                help='Extract ROIs to this directory')
    parser.add_argument('-l', '--limit', default=-1, type=int, help='Only process the first <limit> files in the directory.')
    args = parser.parse_args()
    files = sorted(glob.glob(os.path.join(args.data_dir, '*.*')))
    if args.limit >= 0:
        files = files[0:args.limit]

    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger("evaluate_mrz")

    tic = time.time()
    pool = multiprocessing.Pool(args.jobs)
    log.info("Preparing computation for %d files from %s" % (len(files), args.data_dir))
    log.info("Running %d workers" % args.jobs)
    results = []
    save_roi = args.roi_dir is not None
    for d in [args.success_dir, args.fail_dir, args.roi_dir]:
        if d is not None and not os.path.isdir(d):
            os.mkdir(d)

    def valid_score(mrz):
        return 0 if mrz is None else mrz.valid_score

    def score_change_type(filename, mrz):
        try:
            new_score = mrz.valid_score if mrz is not None else 0
            old_score = int(os.path.basename(filename).split('_')[0])
            schange = new_score - old_score
            return '=' if schange == 0 else ('>' if schange > 0 else '<')
        except Exception:
            return '?'

    method_stats = Counter()

    for result in pool.imap_unordered(process_file, [(f, save_roi) for f in files]):
        filename, mrz, walltime = result
        results.append(result)
        log.info("Processed %s in %0.2fs (score %d) [%s]" % (os.path.basename(filename), walltime, valid_score(mrz), score_change_type(filename, mrz)))
        log.debug("\t%s" % str(mrz))

        vs = valid_score(mrz)
        if args.success_dir is not None and vs > 0:
            shutil.copyfile(filename, os.path.join(args.success_dir, '%d_%s' % (vs, os.path.basename(filename))))
        if args.fail_dir is not None and vs == 0:
            shutil.copyfile(filename, os.path.join(args.fail_dir, '%d_%s' % (vs, os.path.basename(filename))))
        if args.roi_dir is not None and mrz is not None and 'roi' in mrz.aux:
            roi_fn = '%d_roi_%s.png' % (vs, os.path.basename(filename))
            io.imsave(os.path.join(args.roi_dir, roi_fn), mrz.aux['roi'])

        if vs > 0 and 'method' in mrz.aux:
            method_stats[mrz.aux['method']] += 1

    num_files = len(results)
    score_changes = [score_change_type(fn, mrz) for fn, mrz, wt in results]
    scores = [valid_score(mrz) for fn, mrz, wt in results]
    num_perfect = scores.count(100)
    num_invalid = scores.count(0)
    total_score = sum(scores)
    total_computation_walltime = sum([wt for fn, mrz, wt in results])
    total_walltime = time.time() - tic
    log.info("Completed")
    print("Walltime:          %0.2fs" % total_walltime)
    print("Compute walltime:  %0.2fs" % total_computation_walltime)
    print("Processed files:   %d" % num_files)
    print("Perfect parses:    %d" % num_perfect)
    print("Invalid parses:    %d" % num_invalid)
    print("Improved parses:   %d" % len(filter(lambda x: x == '>', score_changes)))
    print("Worsened parses:   %d" % len(filter(lambda x: x == '<', score_changes)))
    print("Total score:       %d" % total_score)
    print("Mean score:        %0.2f" % (float(total_score)/num_files))
    print("Mean compute time: %0.2fs" % (total_computation_walltime/num_files))
    print("Methods used:")
    for stat in method_stats.most_common():
        print("  %s: %d" % stat)

def mrz():
    """
    Command-line script for extracting MRZ from a given image
    """
    parser = argparse.ArgumentParser(description='Run the MRZ OCR recognition algorithm on the given image.')
    parser.add_argument('filename')
    parser.add_argument('--json', action='store_true', help='Produce JSON (rather than tabular) output')
    parser.add_argument('-r', '--save-roi', default=None,
                        help='Output the region of the image that is detected to contain the MRZ to the given png file')
    parser.add_argument('--version', action='version', version='PassportEye MRZ v%s' % passporteye.__version__)
    args = parser.parse_args()

    filename, mrz, walltime = process_file((args.filename, args.save_roi is not None))
    d = mrz.to_dict() if mrz is not None else {'mrz_type': None, 'valid': False, 'valid_score': 0}
    d['walltime'] = walltime
    d['filename'] = filename

    if args.save_roi is not None and mrz is not None and 'roi' in mrz.aux:
        io.imsave(args.save_roi, mrz.aux['roi'])

    if not args.json:
        for k in d:
            print("%s\t%s" % (k, str(d[k])))
    else:
        print(json.dumps(d, indent=2))


'''
PassportEye::Util: Geometry & math utilities

Author: Konstantin Tretyakov
License: MIT
'''
import numpy as np
from sklearn.decomposition import PCA
from matplotlib import pyplot as plt
from matplotlib import patches
from skimage import transform


class RotatedBox(object):
    """
    RotatedBox represents a rectangular box centered at (cx,cy) with dimensions width x height,
    rotated by angle radians counterclockwise.

    >>> RotatedBox.from_points([[0,0], [2,1], [0,1], [2,0]])
    RotatedBox(cx=1.0, cy=0.5, width=2.0, height=1.0, angle=0.0)
    """

    def __init__(self, center, width, height, angle, points=None):
        """Creates a new RotatedBox.

        :param points: This parameter may be used to indicate the set of points used to create the box.
        """
        self.center = np.asfarray(center)
        self.width = width
        self.height = height
        self.angle = angle
        self.points = points

    def __repr__(self):
        return "RotatedBox(cx={0}, cy={1}, width={2}, height={3}, angle={4})".format(self.cx, self.cy, self.width, self.height, self.angle)

    @property
    def cx(self):
        return self.center[0]

    @property
    def cy(self):
        return self.center[1]

    @property
    def area(self):
        return self.width * self.height

    def approx_equal(self, center, width, height, angle, tol=1e-6):
        "Method mainly useful for testing"
        return abs(self.cx - center[0]) < tol and abs(self.cy - center[1]) < tol and abs(self.width - width) < tol and \
               abs(self.height - height) < tol and abs(self.angle - angle) < tol

    def rotated(self, rotation_center, angle):
        """Returns a RotatedBox that is obtained by rotating this box around a given center by a given angle.

        >>> assert RotatedBox([2, 2], 2, 1, 0.1).rotated([1, 1], np.pi/2).approx_equal([0, 2], 2, 1, np.pi/2+0.1)
        """
        rot = np.array([[np.cos(angle), np.sin(angle)], [-np.sin(angle), np.cos(angle)]])
        t = np.asfarray(rotation_center)
        new_c = np.dot(rot.T, (self.center - t)) + t
        return RotatedBox(new_c, self.width, self.height, (self.angle+angle) % (np.pi*2))

    def as_poly(self, margin_width=0, margin_height=0):
        """Converts this box to a polygon, i.e. 4x2 array, representing the four corners starting from lower left to upper left counterclockwise.

        :param margin_width: The additional "margin" that will be added to the box along its width dimension (from both sides) before conversion.
        :param margin_height: The additional "margin" that will be added to the box along its height dimension (from both sides) before conversion.

        >>> RotatedBox([0, 0], 4, 2, 0).as_poly()
        array([[-2., -1.],
               [ 2., -1.],
               [ 2.,  1.],
               [-2.,  1.]])
        >>> RotatedBox([0, 0], 4, 2, np.pi/4).as_poly()
        array([[-0.707..., -2.121...],
               [ 2.121...,  0.707...],
               [ 0.707...,  2.121...],
               [-2.121..., -0.707...]])
        >>> RotatedBox([0, 0], 4, 2, np.pi/2).as_poly()
        array([[ 1., -2.],
               [ 1.,  2.],
               [-1.,  2.],
               [-1., -2.]])
        >>> RotatedBox([0, 0], 0, 0, np.pi/2).as_poly(2, 1)
        array([[ 1., -2.],
               [ 1.,  2.],
               [-1.,  2.],
               [-1., -2.]])
        """
        v_hor = (self.width/2 + margin_width)*np.array([np.cos(self.angle), np.sin(self.angle)])
        v_vert = (self.height/2 + margin_height)*np.array([-np.sin(self.angle), np.cos(self.angle)])
        c = np.array([self.cx, self.cy])
        return np.vstack([c - v_hor - v_vert, c + v_hor - v_vert, c + v_hor + v_vert, c - v_hor + v_vert])

    def plot(self, mode='image', ax=None, **kwargs):
        """Visualize the box on a matplotlib plot.
        :param mode: How should the box coordinates and angle be interpreted.
            - mode `'image'` corresponds to the situation where x coordinate of the box
              denotes the "row of an image" (ie. the Y coordinate of the plot, arranged downwards)
              and y coordinate of the box corresponds to the "column of an image",
              (ie X coordinate of the plot). In other words, box's x goes downwards and y - rightwards.
            - mode `'math'` corresponds to the "mathematics" situation where box's x and y correspond to the X and Y axes of the plot.
        :param ax: the matplotlib axis to draw on. If unspecified, the current axis is used.
        :param kwargs: arguments passed to the matplotlib's `Polygon` patch object. By default, fill is set to False, color to red and lw to 2.
        :return: The created Polygon object.
        """
        ax = ax or plt.gca()
        poly = self.as_poly()
        if mode == 'image':
            poly = poly[:,[1,0]]
        kwargs.setdefault('fill', False)
        kwargs.setdefault('color', 'r')
        kwargs.setdefault('lw', 2)
        p = patches.Polygon(poly, **kwargs)
        ax.add_patch(p)
        return p

    def extract_from_image(self, img, scale=1.0, margin_width=5, margin_height=5):
        """Extracts the contents of this box from a given image.
        For that the image is "unrotated" by the appropriate angle, and the corresponding part is extracted from it.

        Returns an image with dimensions height*scale x width*scale.
        Note that the box coordinates are interpreted as "image coordinates" (i.e. x is row and y is column),
        and box angle is considered to be relative to the vertical (i.e. np.pi/2 is "normal orientation")

        :param img: a numpy ndarray suitable for image processing via skimage.
        :param scale: the RotatedBox is scaled by this value before performing the extraction.
            This is necessary when, for example, the location of a particular feature is determined using a smaller image,
            yet then the corresponding area needs to be extracted from the original, larger image.
            The scale parameter in this case should be width_of_larger_image/width_of_smaller_image.
        :param margin_width: The margin that should be added to the width dimension of the box from each size.
            This value is given wrt actual box dimensions (i.e. not scaled).
        :param margin_height: The margin that should be added to the height dimension of the box from each side.
        :return: a numpy ndarray, corresponding to the extracted region (aligned straight).

        TODO: This could be made more efficient if we avoid rotating the full image and cut out the ROI from it beforehand.
        """
        rotate_by = (np.pi/2 - self.angle)*180/np.pi
        img_rotated = transform.rotate(img, angle=rotate_by, center=[self.center[1]*scale, self.center[0]*scale], resize=True)
        # The resizeable transform will shift the resulting image somewhat wrt original coordinates.
        # When we cut out the box we will compensate for this shift.
        shift_c, shift_r = self._compensate_rotation_shift(img, scale)

        r1 = max(int((self.center[0] - self.height/2 - margin_height)*scale - shift_r), 0)
        r2 = int((self.center[0] + self.height/2 + margin_height)*scale - shift_r)
        c1 = max(int((self.center[1] - self.width/2 - margin_width)*scale - shift_c), 0)
        c2 = int((self.center[1] + self.width/2 + margin_width)*scale - shift_c)
        return img_rotated[r1:r2, c1:c2]

    def _compensate_rotation_shift(self, img, scale):
        """This is an auxiliary method used by extract_from_image.
        It is needed due to particular specifics of the skimage.transform.rotate implementation.
        Namely, when you use rotate(... , resize=True), the rotated image is rotated and shifted by certain amount.
        Thus when we need to cut out the box from the image, we need to account for this shift.
        We do this by repeating the computation from skimage.transform.rotate here.

        TODO: This makes the code uncomfortably coupled to SKImage (e.g. this logic is appropriate for skimage 0.12.1, but not for 0.11,
        and no one knows what happens in later versions). A solution would be to use skimage.transform.warp with custom settings, but we can think of it later.
        """
        ctr = np.asarray([self.center[1]*scale, self.center[0]*scale])
        tform1 = transform.SimilarityTransform(translation=ctr)
        tform2 = transform.SimilarityTransform(rotation=np.pi/2 - self.angle)
        tform3 = transform.SimilarityTransform(translation=-ctr)
        tform = tform3 + tform2 + tform1

        rows, cols = img.shape[0], img.shape[1]
        corners = np.array([
            [0, 0],
            [0, rows - 1],
            [cols - 1, rows - 1],
            [cols - 1, 0]
        ])
        corners = tform.inverse(corners)
        minc = corners[:, 0].min()
        minr = corners[:, 1].min()
        maxc = corners[:, 0].max()
        maxr = corners[:, 1].max()

        # SKImage 0.11 version
        out_rows = maxr - minr + 1
        out_cols = maxc - minc + 1

        # fit output image in new shape
        return ((cols - out_cols) / 2., (rows - out_rows) / 2.)


    @staticmethod
    def from_points(points, box_type='bb'):
        """
        Interpret a given point cloud as a RotatedBox, using PCA to determine the potential orientation (the longest component becomes width)
        This is basically an approximate version of a min-area-rectangle algorithm.
        TODO: Test whether using a true min-area-rectangle algorithm would be more precise or faster.

        :param points: An n x 2 numpy array of coordinates.
        :param box_type: The kind of method used to estimate the "box".
            Possible values:
                - `'bb'`, denoting the "bounding box" approach (min/max coordinates of the points correspond to box limits)
                - `'mrz`, denoting a slightly modified technique, suited for MRZ zone detection from contour images.
                          Here the assumption is that the upper and lower bounds of the box are better estimated as the
                          10% and 90% quantile of the corresponding coordinates (rather than 0% and 100%, i.e. min and max).
                          This helps against accidental noise in the contour.
                          The `'mrz'` correction is only applied when there are at least 10 points in the set.
        :returns: a RotatedBox, bounding the given set of points, oriented according to the principal components.

        >>> RotatedBox.from_points([[0,0]])
        RotatedBox(cx=0.0, cy=0.0, width=0.0, height=0.0, angle=0.0)
        >>> assert RotatedBox.from_points([[0,0], [1,1], [2,2]]).approx_equal([1, 1], np.sqrt(8), 0, np.pi/4)
        >>> assert RotatedBox.from_points([[0,0], [1,1], [0,1], [1,0]]).approx_equal([0.5, 0.5], 1, 1, 0.0) # The angle is rather arbitrary here
        >>> assert RotatedBox.from_points([[0,0], [2,1], [0,1], [2,0]]).approx_equal([1, 0.5], 2, 1, 0)
        >>> assert RotatedBox.from_points([[0,0], [2,4], [0,4], [2,0]]).approx_equal([1, 2], 4, 2, np.pi/2)
        >>> assert RotatedBox.from_points([[0,0], [1,1.5], [2,0]]).approx_equal([1, 0.75], 2, 1.5, 0)
        >>> assert RotatedBox.from_points([[0,0], [0,1], [1,1]]).approx_equal([0.25, 0.75], np.sqrt(2), np.sqrt(2)/2, np.pi/4)
        """
        points = np.asfarray(points)
        if points.shape[0] == 1:
            return RotatedBox(points[0], width=0.0, height=0.0, angle=0.0, points=points)

        m = PCA(2).fit(points)
        # Find the angle
        angle = (np.arctan2(m.components_[0,1], m.components_[0,0]) % np.pi)
        if abs(angle - np.pi) < angle:
            # Here the angle is always between -pi and pi
            # If the principal component happened to be oriented so that the angle happens to be > pi/2 by absolute value,
            # we flip the direction
            angle = angle - np.pi if angle > 0 else angle + np.pi
        points_transformed = m.transform(points)
        ll = np.min(points_transformed, 0)
        ur = np.max(points_transformed, 0)
        wh = ur - ll

        # Now compute and return the bounding box
        if box_type == 'bb' or (box_type == 'mrz' and points.shape[0] < 10):
            # We know that if we rotate the points around m.mean_, we get a box with bounds ur and ll
            # The center of this box is (ur+ll)/2 + mean, which is not the same as the mean,
            # hence to get the center of the original box we need to "unrotate" this box back.
            return RotatedBox(np.dot(m.components_.T, (ll+ur)/2) + m.mean_, width=wh[0], height=wh[1], angle=angle, points=points)
        elif box_type == 'mrz':
            # When working with MRZ detection from contours, we may have minor "bumps" in the contour,
            # that should be ignored at least along the long ("horizontal") side.
            # To do that, we will use 10% and 90% quantiles as the bounds of the box instead of the max and min.
            # We drop all points which lie beyond and simply repeat the estimation (now 'bb-style') without them.
            h_coord = sorted(points_transformed[:,1])
            n = len(h_coord)
            bottom, top = h_coord[n/10], h_coord[n*9/10]
            valid_points = np.logical_and(points_transformed[:,1]>=bottom, points_transformed[:,1]<=top)
            rb = RotatedBox.from_points(points[valid_points, :], 'bb')
            rb.points = points
            return rb
        else:
            raise ValueError("Unknown parameter value: box_type=%s" % box_type)




'''
PassportEye::MRZ: Machine-readable zone extraction and parsing.
Image processing for MRZ extraction.

Author: Konstantin Tretyakov
License: MIT
'''
from skimage import transform, io, morphology, filters, measure
import numpy as np
import tempfile, os
from ..util.pdf import extract_first_jpeg_in_pdf
from ..util.pipeline import Pipeline
from ..util.geometry import RotatedBox
from ..util.ocr import ocr
from .text import MRZ


class Loader(object):
    """Loads `filename` to `img`."""

    __depends__ = []
    __provides__ = ['img']

    def __init__(self, filename, as_grey=True, pdf_aware=True):
        self.filename = filename
        self.as_grey = as_grey
        self.pdf_aware = pdf_aware

    def _imread(self, filename):
        """Proxy to skimage.io.imread with some fixes."""
        img = io.imread(filename, as_grey=self.as_grey)
        if img is not None and len(img.shape) != 2:
            # The PIL plugin somewhy fails to load some images
            img = io.imread(filename, as_grey=self.as_grey, plugin='matplotlib')
        return img

    def __call__(self):
        if self.pdf_aware and self.filename.lower().endswith('.pdf'):
            with open(self.filename, 'rb') as f:
                img_data = extract_first_jpeg_in_pdf(f)
            if img_data is None:
                return None
            else:
                fd, fname = tempfile.mkstemp(prefix='pythoneye_', suffix='.jpg')
                try:
                    with open(fname, 'wb') as f:
                        f.write(img_data)
                    return self._imread(fname)
                except:
                    return None
                finally:
                    os.close(fd)
                    os.remove(fname)
        else:
            return self._imread(self.filename)


class Scaler(object):
    """Scales `image` down to `img_scaled` so that its width is at most 250."""

    __depends__ = ['img']
    __provides__ = ['img_small', 'scale_factor']

    def __init__(self, max_width=250):
        self.max_width = max_width

    def __call__(self, img):
        scale_factor = self.max_width/float(img.shape[1])
        if scale_factor <= 1:
            img_small = transform.rescale(img, scale_factor)
        else:
            scale_factor = 1.0
            img_small = img
        return img_small, scale_factor


class BooneTransform(object):
    """Processes `img_small` according to Hans Boone's method
    (http://www.pyimagesearch.com/2015/11/30/detecting-machine-readable-zones-in-passport-images/)
    Outputs a `img_binary` - a result of threshold_otsu(closing(sobel(black_tophat(img_small)))"""

    __depends__ = ['img_small']
    __provides__ = ['img_binary']

    def __init__(self, square_size=5):
        self.square_size = 5

    def __call__(self, img_small):
        m = morphology.square(self.square_size)
        img_th = morphology.black_tophat(img_small, m)
        img_sob = abs(filters.sobel_v(img_th))
        img_closed = morphology.closing(img_sob, m)
        threshold = filters.threshold_otsu(img_closed)
        return img_closed > threshold


class MRZBoxLocator(object):
    """Extracts putative MRZs as RotatedBox instances from the contours of `img_binary`"""

    __depends__ = ['img_binary']
    __provides__ = ['boxes']

    def __init__(self, max_boxes=4, min_points_in_contour=50, min_area=500, min_box_aspect=5, angle_tol=0.1, lineskip_tol=1.5, box_type='bb'):
        self.max_boxes = max_boxes
        self.min_points_in_contour = min_points_in_contour
        self.min_area = min_area
        self.min_box_aspect = min_box_aspect
        self.angle_tol = angle_tol
        self.lineskip_tol = lineskip_tol
        self.box_type = box_type

    def __call__(self, img_binary):
        cs = measure.find_contours(img_binary, 0.5)

        # Collect contours into RotatedBoxes
        results = []
        for c in cs:
            # Now examine the bounding box. If it is too small, we ignore the contour
            ll, ur = np.min(c, 0), np.max(c, 0)
            wh = ur - ll
            if wh[0]*wh[1] < self.min_area: continue

            # Finally, construct the rotatedbox. If its aspect ratio is too small, we ignore it
            rb = RotatedBox.from_points(c, self.box_type)
            if rb.height == 0 or rb.width/rb.height < self.min_box_aspect: continue

            # All tests fine, add to the list
            results.append(rb)

        # Next sort and leave only max_boxes largest boxes by area
        results.sort(lambda x,y: 1 if x.area < y.area else -1)
        return self._merge_boxes(results[0:self.max_boxes])

    def _are_aligned_angles(self, b1, b2):
        "Are two boxes aligned according to their angle?"
        return abs(b1 - b2) <= self.angle_tol or abs(np.pi - abs(b1 - b2)) <= self.angle_tol

    def _are_nearby_parallel_boxes(self, b1, b2):
        "Are two boxes nearby, parallel, and similar in width?"
        if not self._are_aligned_angles(b1.angle, b2.angle): return False
        # Otherwise pick the smaller angle and see whether the two boxes are close according to the "up" direction wrt that angle
        angle = min(b1.angle, b2.angle)
        return abs(np.dot(b1.center - b2.center, [-np.sin(angle), np.cos(angle)])) < self.lineskip_tol*(b1.height + b2.height) and \
               (b1.width > 0) and (b2.width > 0) and (0.5 < b1.width/b2.width < 2.0)

    def _merge_any_two_boxes(self, box_list):
        """Given a list of boxes, finds two nearby parallel ones and merges them. Returns false if none found."""
        for i in range(len(box_list)):
            for j in range(i+1,len(box_list)):
                if self._are_nearby_parallel_boxes(box_list[i], box_list[j]):
                    # Remove the two boxes from the list, add a new one
                    a, b = box_list[i], box_list[j]
                    merged_points = np.vstack([a.points, b.points])
                    merged_box =  RotatedBox.from_points(merged_points, self.box_type)
                    if merged_box.width/merged_box.height >= self.min_box_aspect:
                        box_list.remove(a)
                        box_list.remove(b)
                        box_list.append(merged_box)
                        return True
        return False

    def _merge_boxes(self, box_list):
        """Mergest nearby parallel boxes in the given list."""
        while self._merge_any_two_boxes(box_list):
            pass
        return box_list


class FindFirstValidMRZ(object):
    """Iterates over boxes found by MRZBoxLocator, passes them to BoxToMRZ, finds the first valid MRZ
    or the best-scoring MRZ"""

    __provides__ = ['box_idx', 'roi', 'text', 'mrz']
    __depends__ = ['boxes', 'img', 'img_small', 'scale_factor', '__data__']

    def __init__(self, use_original_image=True):
        self.box_to_mrz = BoxToMRZ(use_original_image)

    def __call__(self, boxes, img, img_small, scale_factor, data):
        mrzs = []
        data['__debug__mrz'] = []
        for i, b in enumerate(boxes):
            roi, text, mrz = self.box_to_mrz(b, img, img_small, scale_factor)
            data['__debug__mrz'].append((roi, text, mrz))
            if mrz.valid:
                return i, roi, text, mrz
            elif mrz.valid_score > 0:
                mrzs.append((i, roi, text, mrz))
        if len(mrzs) == 0:
            return None, None, None, None
        else:
            mrzs.sort(cmp = lambda x,y: x[3].valid_score - y[3].valid_score)
            return mrzs[-1]


class BoxToMRZ(object):
    """Extracts ROI from the image, corresponding to a box found by MRZBoxLocator, does OCR and MRZ parsing on this region."""

    __provides__ = ['roi', 'text', 'mrz']
    __depends__ = ['box', 'img', 'img_small', 'scale_factor']

    def __init__(self, use_original_image=True):
        """
        :param use_original_image: when True, the ROI is extracted from img, otherwise from img_small
        """
        self.use_original_image = use_original_image

    def __call__(self, box, img, img_small, scale_factor):
        img = img if self.use_original_image else img_small
        scale = 1.0/scale_factor if self.use_original_image else 1.0

        # If the box's angle is np.pi/2 +- 0.01, we shall round it to np.pi/2:
        # this way image extraction is fast and introduces no distortions.
        # and this may be more important than being perfectly straight
        # similar for 0 angle
        if abs(abs(box.angle) - np.pi/2) <= 0.01:
            box.angle = np.pi/2
        if abs(box.angle) <= 0.01:
            box.angle = 0.0

        roi = box.extract_from_image(img, scale)
        text = ocr(roi)

        if '>>' in text or ('>' in text and '<' not in text):
            # Most probably we need to reverse the ROI
            roi = roi[::-1,::-1]
            text = ocr(roi)

        if not '<' in text:
            # Assume this is unrecoverable and stop here (TODO: this may be premature, although it saves time on useless stuff)
            return roi, text, MRZ.from_ocr(text)

        mrz = MRZ.from_ocr(text)
        mrz.aux['method'] = 'direct'

        # Now try improving the result via hacks
        if not mrz.valid:
            text, mrz = self._try_larger_image(roi, text, mrz)

        # Sometimes the filter used for enlargement is important!
        if not mrz.valid:
            text, mrz = self._try_larger_image(roi, text, mrz, 1)

        if not mrz.valid:
            text, mrz = self._try_black_tophat(roi, text, mrz)


        return roi, text, mrz

    def _try_larger_image(self, roi, cur_text, cur_mrz, filter_order=3):
        """Attempts to improve the OCR result by scaling the image. If the new mrz is better, returns it, otherwise returns
        the old mrz."""
        if roi.shape[1] <= 700:
            scale_by = int(1050.0/roi.shape[1] + 0.5)
            roi_lg = transform.rescale(roi, scale_by, order=filter_order)
            new_text = ocr(roi_lg)
            new_mrz = MRZ.from_ocr(new_text)
            new_mrz.aux['method'] = 'rescaled(%d)' % filter_order
            if new_mrz.valid_score > cur_mrz.valid_score:
                cur_mrz = new_mrz
                cur_text = new_text
        return cur_text, cur_mrz

    def _try_black_tophat(self, roi, cur_text, cur_mrz):
        roi_b = morphology.black_tophat(roi, morphology.disk(5))
        new_text = ocr(roi_b)  # There are some examples where this line basically hangs for an undetermined amount of time.
        new_mrz = MRZ.from_ocr(new_text)
        if new_mrz.valid_score > cur_mrz.valid_score:
            new_mrz.aux['method'] = 'black_tophat'
            cur_text, cur_mrz = new_text, new_mrz

        new_text, new_mrz = self._try_larger_image(roi_b, cur_text, cur_mrz)
        if new_mrz.valid_score > cur_mrz.valid_score:
            new_mrz.aux['method'] = 'black_tophat(rescaled(3))'
            cur_text, cur_mrz = new_text, new_mrz

        return cur_text, cur_mrz


class TryOtherMaxWidth(object):
    """
    If mrz was not found so far in the current pipeline,
    changes the max_width parameter of the scaler to 1000 and reruns the pipeline again.
    """

    __provides__ = ['mrz_final']
    __depends__ = ['mrz', '__pipeline__']

    def __init__(self, other_max_width=1000):
        self.other_max_width = other_max_width

    def __call__(self, mrz, __pipeline__):
        # We'll only try this if we see that img_binary.mean() is very small or img.mean() is very large (i.e. image is mostly white).
        if mrz is None and (__pipeline__['img_binary'].mean() < 0.01 or __pipeline__['img'].mean() > 0.95):
            __pipeline__.replace_component('scaler', Scaler(self.other_max_width))
            new_mrz = __pipeline__['mrz']
            new_mrz.aux['method'] = new_mrz.aux['method'] + '|max_width(%d)' % self.other_max_width
            mrz = new_mrz
        return mrz


class MRZPipeline(Pipeline):
    """This is the "currently best-performing" pipeline for parsing MRZ from a given image file."""

    def __init__(self, filename):
        super(MRZPipeline, self).__init__()
        self.version = '1.0'  # In principle we might have different pipelines in use, so possible backward compatibility is an issue
        self.filename = filename
        self.add_component('loader', Loader(filename))
        self.add_component('scaler', Scaler())
        self.add_component('boone', BooneTransform())
        self.add_component('box_locator', MRZBoxLocator())
        self.add_component('mrz', FindFirstValidMRZ())
        self.add_component('other_max_width', TryOtherMaxWidth())

    @property
    def result(self):
        return self['mrz_final']


def read_mrz(filename, save_roi=False):
    """The main interface function to this module, encapsulating the recognition pipeline.
       Given an image filename, runs MRZPipeline on it, returning the parsed MRZ object.

    :param save_roi: when this is True, the .aux['roi'] field will contain the Region of Interest where the MRZ was parsed from.
    """
    p = MRZPipeline(filename)
    mrz = p.result

    if mrz is not None:
        if save_roi: mrz.aux['roi'] = p['roi']
    return mrz



'''
PassportEye::Util: Pipeline utility

Author: Konstantin Tretyakov
License: MIT
'''


class Pipeline(object):
    """
    The computation performed in order to extract the information from an image
    is essentially a list of steps of various complexity, where each step uses
    results of the previous steps and introduces its own results.

    Although this is nothing more than a standard sequential program, it seems to be somewhat
    more convenient sometimes to regard it as a "pipeline" consisting of pluggable "components",
    where each component explicitly specifies what inputs it needs and what outputs it produces,
    and the workflow engine wires up the inputs to the outputs.

    This class offers provides a simple implementation of such a pipeline.
    It keeps track of a dictionary of values that were already computed, a dictionary of
    "components" which know how to compute other values, and routes item accesses to computations automatically.

    >>> a = Pipeline()
    >>> a.add_component('1', lambda: 1, ['a'], [])
    >>> a.add_component('2', lambda: 2, ['b'], [])
    >>> a.add_component('s,d', lambda x,y: (x+y, x-y), ['c', 'd'], ['a', 'b'])
    >>> a.add_component('sd', lambda x,y: (x+y, x-y), ['e'], ['a', 'b'])
    >>> a['c']
    3
    >>> a['d']
    -1
    >>> a['e']
    (3, -1)
    >>> a.replace_component('1', lambda: 2, ['a'], [])
    >>> a['e']
    (4, 0)
    >>> a['d']
    0
    """

    def __init__(self):
        self.data = dict()        # Maps key -> data item.
        self.components = dict()  # Maps name -> component
        self.provides = dict()    # Component name -> provides list
        self.depends = dict()     # Component name -> depends list
        self.whoprovides = dict() # key -> component name
        self.data['__data__'] = self.data
        self.data['__pipeline__'] = self

    def add_component(self, name, callable, provides=None, depends=None):
        """
        Add a given callable to a list of components. The provides and depends are lists of strings, specifying what
        keys the component computes and what keys it requires to be present. If those are not given, the callable must
        have fields __provides__ and __depends__.
        """
        provides = provides or getattr(callable, '__provides__', [])
        depends = depends or getattr(callable, '__depends__', [])
        for p in provides:
            if p in self.whoprovides:
                raise Exception("There is already a component that provides %s" % p)
        self.provides[name] = provides
        self.depends[name] = depends
        self.components[name] = callable
        for p in provides:
            self.whoprovides[p] = name

    def remove_component(self, name):
        """Removes an existing component with a given name, invalidating all the values computed by
        the previous component."""
        if name not in self.components:
            raise Exception("No component named %s" % name)
        del self.components[name]
        del self.depends[name]
        for p in self.provides[name]:
            del self.whoprovides[p]
            self.invalidate(p)
        del self.provides[name]

    def replace_component(self, name, callable, provides=None, depends=None):
        """Changes an existing component with a given name, invalidating all the values computed by
        the previous component and its successors."""
        self.remove_component(name)
        self.add_component(name, callable, provides, depends)

    def invalidate(self, key):
        """Remove the given data item along with all items that depend on it in the graph."""
        if key not in self.data:
            return
        del self.data[key]

        # Find all components that used it and invalidate their results
        for cname in self.components:
            if key in self.depends[cname]:
                for downstream_key in self.provides[cname]:
                    self.invalidate(downstream_key)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        self._compute(key)
        return self.data[key]

    def _compute(self, key):
        if key not in self.data:
            cname = self.whoprovides[key]
            for d in self.depends[cname]:
                self._compute(d)
            inputs = [self.data[d] for d in self.depends[cname]]
            results = self.components[cname](*inputs)
            if len(self.provides[cname]) == 1:
                self.data[self.provides[cname][0]] = results
            else:
                for k, v in zip(self.provides[cname], results):
                    self.data[k] = v

